#include "boost/numeric/ublas/vector_proxy.hpp"
#include "boost/numeric/ublas/matrix_proxy.hpp"
#include "hpp/localstepper/solver-basic.hh"
#include "hpp/hik/solver-basic.hh"
#include "hpp/localstepper/task/task.hh"
#include "hpp/localstepper/centipede.hh"
#include "hpp/gik/tools.hh"

#define M4_IJ MAL_S4x4_MATRIX_ACCESS_I_J
#define M3_IJ MAL_S3x3_MATRIX_ACCESS_I_J

using namespace boost::numeric::ublas;

ChppLosSolverBasic::ChppLosSolverBasic ( ChppLosCentipede& inC )
{
    attCentipede = &inC;

    attHikSolver = new ChppHikSolverBasic ( attCentipede->numberDof() );

    attHikSolver->setMask ( scalar_vector<double> ( attCentipede->numberDof(), 1 ) );

    attSolution.resize ( attCentipede->numberDof() );
    attUb.resize ( attCentipede->numberDof() );
    attLb.resize ( attCentipede->numberDof() );
}


ChppLosSolverBasic::~ChppLosSolverBasic()
{
    delete attHikSolver;
}

void ChppLosSolverBasic::tasks ( const std::vector<ChppLosTask*>& inTasks )
{
    attTasks = inTasks;
}

void ChppLosSolverBasic::constraints ( const std::vector<ChppLosTask*>& inConstraints )
{
    attConstraints = inConstraints;
}

bool ChppLosSolverBasic::setMask ( const vectorN& inVector )
{
    if ( attHikSolver->numberParams() != attCentipede->numberDof() )
        resize();
    
    if ( inVector.size() != attCentipede->numberDof() )
    {
        std::cout << "ChppLosSolverBasic::setMask -Error bad vector size"<< std::endl;
        return false;
    }

    attHikSolver->setMask ( inVector );
    return true;
}

void ChppLosSolverBasic::resize()
{
    delete attHikSolver;
    attHikSolver = new ChppHikSolverBasic ( attCentipede->numberDof() );
    attUb.resize ( attCentipede->numberDof() );
    attLb.resize ( attCentipede->numberDof() );
}

bool ChppLosSolverBasic::solve ( vectorN& veloSolution )
{
    int retV;
    unsigned int i;

    if ( veloSolution.size() != attCentipede->numberDof() )
    {
        std::cout << "ChppLosSolverBasic::solve Bad size of initial guess" << std::endl;
        return false;
    }

    if ( attHikSolver->numberParams() != attCentipede->numberDof() )
        resize();

    //Set parameter bounds
    attHikSolver->clearBounds();
    attCentipede->getUpperVelocityBounds ( attUb );
    attCentipede->getLowerVelocityBounds ( attLb );
    for ( i=0;i<attCentipede->numberDof();i++ )
    {
        attHikSolver->upperBound ( i, attUb ( i ) );
        attHikSolver->lowerBound ( i, attLb ( i ) );
    }

    //Set initial feasible set
    loadConstraintsInSolver();


    for ( i = 0; i< attTasks.size(); i++ )
    {
        attHikSolver->clearTasks();
        attTasks[i]->computeDifferentialSystem();
        attHikSolver->addTask ( attTasks[i]->differentialSystem(), attTasks[i]->isEqualityTask() );
        retV = attHikSolver->solve ( veloSolution, attTasks[i]->dampingFactor() );
        if ( retV !=0 )
        {
            std::cout << "retV = " << retV << " at task ranked "<< i<< std::endl;
            return false;
        }

        if ( i!= attTasks.size()-1 )
            attHikSolver->updateFeasibleSet();
    }
    veloSolution = attHikSolver->solution();

    return true;
}


void ChppLosSolverBasic::loadConstraintsInSolver()
{
    attHikSolver->removePreConstraints();

    for ( unsigned int i=0;i<attConstraints.size();i++ )
    {
        attConstraints[i]->computeDifferentialSystem();
        if ( attConstraints[i]->isEqualityTask() )
            attHikSolver->addEqPreConstraint ( attConstraints[i]->differentialSystem() );
        else
            attHikSolver->addIneqPreConstraint ( attConstraints[i]->differentialSystem() );
    }
    attHikSolver->initFeasibleSet();
}

