#include "boost/numeric/ublas/vector_proxy.hpp"
#include "boost/numeric/ublas/matrix_proxy.hpp"
#include "hpp/localstepper/solver.hh"
#include "hpp/localstepper/solver-basic.hh"
#include "hpp/localstepper/task/task.hh"
#include "hpp/localstepper/centipede.hh"
#include "hpp/localstepper/link.hh"
#include "hpp/gik/robot/standing-robot.hh"
#include "robotDynamics/jrlHumanoidDynamicRobot.h"
#include "hpp/gik/tools.hh"
#define M4_IJ MAL_S4x4_MATRIX_ACCESS_I_J
#define M3_IJ MAL_S3x3_MATRIX_ACCESS_I_J

using namespace boost::numeric::ublas;

ChppLosSolver::ChppLosSolver ( ChppLosCentipede& inC )
{
    attCentipede = &inC;

    attLosSolver = new ChppLosSolverBasic ( inC );

    attActivatedJoints = scalar_vector<double> ( attCentipede->numberDof(), 1 );
    attLosSolver->setMask ( attActivatedJoints );

    attSolveThreshold = 1e-3;
    attProgressThreshold = -1e-4;

    veloSolution.resize ( attCentipede->numberDof() );
    attSolution.resize ( attCentipede->numberDof() );
    attMaxIter = 10;
}


ChppLosSolver::~ChppLosSolver()
{
    delete attLosSolver;
}

void ChppLosSolver::tasks ( const std::vector<ChppLosTask*>& inTasks )
{
    attTasks = inTasks;
}

void ChppLosSolver::constraints ( const std::vector<ChppLosTask*>& inConstraints )
{
    attConstraints = inConstraints;
}

bool ChppLosSolver::setMask ( const vectorN& inVector )
{
    if ( inVector.size() != attCentipede->numberDof() )
    {
        std::cout << "ChppLosSolver::setMask -Error bad vector size"<< std::endl;
        return false;
    }

    attActivatedJoints = inVector;
    attLosSolver->setMask ( attActivatedJoints );
    return true;
}

const std::list<vectorN>& ChppLosSolver::solutionMotion() const
{
    return attSolutionMotion;
}
bool ChppLosSolver::solve ()
{
    
    attSolution.resize(attCentipede->numberDof());
    attCentipede->getConfiguration(attSolution);
    veloSolution.resize ( attSolution.size() );
    curConfig.resize ( attSolution.size() );
    
    attSolutionMotion.clear();
    attSolutionMotion.push_back(attSolution);;

    if ( attTasks.empty() )
    {
        std::cout << "ChppLosSolver::solve Nothing to do" << std::endl;
        return true;
    }

    attLosSolver->tasks ( attTasks );
    attLosSolver->constraints ( attConstraints );

    attSolvedTasks.resize(attTasks.size());
    attSolvedTasks.clear();
    
    while ( 1 )
    {
        recordGaps();
        if ( solveIteration() )
        {
            attCentipede->getConfiguration(attSolution);
            attSolutionMotion.push_back(attSolution);
        }
        else
        {
            attCentipede->applyConfiguration ( attSolution );
            break;
        }
        if ( !smallerGaps() )
            break;
    }
    bool retV = isProblemSolved();
    return retV;
}

bool ChppLosSolver::isProblemSolved()
{
    if ( !constraintsSatisfied() )
    {
        //std::cout << "Last config's constraints not satisfied" << std::endl;
        return false;
    }
    for ( unsigned int i = 0;i<attTasks.size();i++ )
    {
        //std::cout << "Task value norm at rank "<<i<< " "<<  norm_2 ( attTasks[i]->currentValue() ) << std::endl; 
        if ( norm_2 ( attTasks[i]->currentValue() ) > attSolveThreshold )
        {
            std::cout << "Task at rank "<< i <<" considered unsolved. Task value norm is "<< norm_2 ( attTasks[i]->currentValue() ) <<std::endl;
            return false;
        }
    }

    return true;
}


bool ChppLosSolver::solveIteration( )
{
    //std::cout << "solve step" << std::endl;
    bool constraintsOK= false;

    unsigned int nHikIter = attMaxIter;


    while ( nHikIter>0 )
    {
        if ( constraintsOK )
            return true;

        veloSolution.clear();

        attLosSolver->solve ( veloSolution );

        attCentipede->getConfiguration(curConfig);
        curConfig.plus_assign(veloSolution);
        attCentipede->applyConfiguration (curConfig);

        constraintsOK = constraintsSatisfied();
        
        if ( !constraintsOK )
        {
            std::cout << "Constraints not satisfied" <<std::endl;
        }

        nHikIter--;
    }

    return constraintsOK;
}


bool ChppLosSolver::constraintsSatisfied()
{
    for ( unsigned int i = 0; i < attConstraints.size(); i++ )
        if ( norm_2 ( attConstraints[i]->currentValue() ) > attSolveThreshold )
        {
            //std::cout << "constraint at rank "<< i <<" not satisfied" << std::endl;
            std::cout << attConstraints[i]->currentValue() << std::endl;
            return false;
        }
    return true;
}

void ChppLosSolver::recordGaps()
{
    attTaskGaps.clear();
    for ( unsigned int i=0; i< attTasks.size();i++ )
    {
        //std::cout << "gap " << i << " " << norm_2 ( attTasks[i]->currentValue() ) << std::endl;
        attTaskGaps.push_back ( norm_2 ( attTasks[i]->currentValue() ) );
    }
}

bool ChppLosSolver::smallerGaps()
{
    std::vector<double> newGaps;
    double change;

    for ( unsigned int i=0; i< attTasks.size();i++ )
    {
        //std::cout << "newgap " << i << " " << norm_2 ( attTasks[i]->currentValue() ) << std::endl;
        newGaps.push_back ( norm_2 ( attTasks[i]->currentValue() ) );
    }

    for ( unsigned int wc =0; wc<attTaskGaps.size();wc++ )
    {
        
        if (attSolvedTasks(wc) == 1)
            continue;
             
        if (norm_2 ( attTasks[wc]->currentValue() )< attSolveThreshold)
        {
            
            bool allPrevSolved = true;
            for ( unsigned int wc2 =0; wc2<wc;wc2++ )
            {
                if (attSolvedTasks(wc2)==0)
                {
                    allPrevSolved = false;
                    break;
                }
            }
            if (allPrevSolved)
                attSolvedTasks(wc) = 1;
            
             return true;
        }
        
        change = newGaps[wc] - attTaskGaps[wc];
        if ( change > 0 )
        {
            std::cout << "Divergence for task at rank "<< wc<< std::endl;
            break;
        }
        else
            if ( change > attProgressThreshold )
                continue;
            else
            {
                attTaskGaps = newGaps;
                return true;
            }
    }
    std::cout << "Local minimum reached." << std::endl;
    return false;
}
