#include "boost/numeric/ublas/vector_proxy.hpp"
#include "boost/numeric/ublas/matrix_proxy.hpp"
#include "hpp/localstepper/task/coplanar-task.hh"
#include "hpp/localstepper/centipede.hh"
#include "hpp/localstepper/link.hh"
#include "hpp/hik/linear-system.hh"
#include "hpp/gik/robot/standing-robot.hh"
#include "robotDynamics/jrlHumanoidDynamicRobot.h"
using namespace boost::numeric::ublas;

#define M3_IJ MAL_S3x3_MATRIX_ACCESS_I_J
#define M4_IJ MAL_S4x4_MATRIX_ACCESS_I_J

void ChppLosCoplanarTask::equivAsymMat3 ( const vector3d& inVec, matrixNxP& outMat )
{
    outMat ( 0,0 ) = 0;
    outMat ( 0,1 ) = -inVec ( 2 );
    outMat ( 0,2 ) = inVec ( 1 );
    outMat ( 1,0 ) = inVec ( 2 );
    outMat ( 1,1 ) = 0;
    outMat ( 1,2 ) = -inVec ( 0 );
    outMat ( 2,0 ) = -inVec ( 1 );
    outMat ( 2,1 ) = inVec ( 0 );
    outMat ( 2,2 ) = 0;
}

void ChppLosCoplanarTask::crossProd3 ( const vector3d& v1, const vector3d& v2, vector3d& retV )
{
    retV ( 0 ) = v1 ( 1 ) *v2 ( 2 ) - v1 ( 2 ) *v2 ( 1 );
    retV ( 1 ) = v1 ( 2 ) *v2 ( 0 ) - v1 ( 0 ) *v2 ( 2 );
    retV ( 2 ) = v1 ( 0 ) *v2 ( 1 ) - v1 ( 1 ) *v2 ( 0 );
}

ChppLosCoplanarTask::ChppLosCoplanarTask ( ChppLosCentipede& inCentipede, ChppLosRobotLink& inRobotLink, CjrlJoint& inJoint, const vector3d& inLocalC, const vector3d& inLocalD, const vector3d& inWorldA,  const vector3d& inWorldNormal ) :ChppLosTaskOnJoint ( inCentipede,inJoint )
{
    attRobotLink = &inRobotLink;
    attA = inWorldA;
    attD = inLocalD;
    attC = inLocalC;
    attNormal = inWorldNormal;

    attConstraint->matrix().resize ( 3,attCentipede->numberDof() );
    attConstraint->matrix().clear();
    attConstraint->vector().resize ( 3 );
    attTempJacobian.resize ( 3, attCentipede->numberDof() );
    attValue.resize ( 1 );
    attMaxVelocity = 1.0;

    attMatNormal.resize ( 3,3 );
    equivAsymMat3 ( attNormal,attMatNormal );
}

void  ChppLosCoplanarTask::localC ( const vector3d& inPoint ) {    attC = inPoint;}
void  ChppLosCoplanarTask::localD ( const vector3d& inPoint ) {    attD = inPoint;}
void  ChppLosCoplanarTask::worldA ( const vector3d& inPoint ) {    attA = inPoint;}
void  ChppLosCoplanarTask::worldNormal ( const vector3d& inPoint )
{
    attNormal = inPoint;
    equivAsymMat3 ( attNormal,attMatNormal );
}
const vector3d& ChppLosCoplanarTask::localC() {return attC;}
const vector3d& ChppLosCoplanarTask::localD() {return attD;}
const vector3d& ChppLosCoplanarTask::worldA() {return attA;}
const vector3d& ChppLosCoplanarTask::worldNormal() {return attNormal;}

void ChppLosCoplanarTask::maxVelocityNorm ( double inMV ) { attMaxVelocity = ( inMV>0.0 ) ?inMV:0.0;}


bool ChppLosCoplanarTask::resize()
{
    if ( ChppLosTask::resize() )
    {
        attTempJacobian.resize ( 3,attCentipede->numberDof() );
        return true;
    }
    return false;
}

void ChppLosCoplanarTask::computeDifferentialSystem()
{
    resize();

    m =attJoint->currentTransformation();
    vector3d absC = m*attC;
    vector3d absD = m*attD;


    vector3d AC = absC - attA;
    vector3d AD = absD - attA;
    vector3d croC,croD,cro3;

    crossProd3 ( AD, attNormal,croD );
    crossProd3 ( AC, attNormal,croC );
    crossProd3 ( croC, croD, cro3 );

    attConstraint->vector() ( 0 ) = cro3 ( 0 );
    attConstraint->vector() ( 1 ) = cro3 ( 1 );
    attConstraint->vector() ( 2 ) = cro3 ( 2 );
    double normcross = MAL_S3_VECTOR_NORM ( cro3 );
    
    if ( normcross > attMaxVelocity )
        attConstraint->vector() = attMaxVelocity*attConstraint->vector() /normcross;

    

    //task jacobian
    matrixNxP matcroC ( 3,3 ), matcroD ( 3,3 ),matcroCun ( 3,3 ), matcroDun(3,3);
    equivAsymMat3 ( croC,matcroC );
    equivAsymMat3 ( croD,matcroD );
    noalias ( matcroCun ) = prod ( matcroC,attMatNormal );
    noalias ( matcroDun ) = prod ( matcroD,attMatNormal );

    
    //Point jacobian
    attTempJacobian.clear();
    attRobotLink->standingRobot().robot()->getPositionJacobian ( *attRobotLink->rootJoint(),*attJoint, attD, attTempJacobian, attRobotLink->rank(), false );
    ChppLosLink* plink = attRobotLink->parent();
    while ( plink!=0 )
    {
        plink->getPositionJacobian ( absD,attTempJacobian,plink->rank() );
        plink = plink->parent();
    }
    noalias(attConstraint->matrix()) =  prod ( matcroCun,attTempJacobian) ;
    
    //Point jacobian
    attTempJacobian.clear();
    attRobotLink->standingRobot().robot()->getPositionJacobian ( *attRobotLink->rootJoint(),*attJoint, attC, attTempJacobian, attRobotLink->rank(), false );
    plink = attRobotLink->parent();
    while ( plink!=0 )
    {
        plink->getPositionJacobian ( absC,attTempJacobian,plink->rank() );
        plink = plink->parent();
    }
    attConstraint->matrix().minus_assign( prod ( matcroDun,attTempJacobian) );
}

const vectorN& ChppLosCoplanarTask::currentValue()
{
    m =attJoint->currentTransformation();
    vector3d absC = m*attC;
    vector3d absD = m*attD;


    vector3d AC = absC - attA;
    vector3d AD = absD - attA;
    vector3d cro1,cro2,cro3;

    crossProd3 ( AD, attNormal,cro1 );
    crossProd3 ( AC,attNormal, cro2 );
    crossProd3 ( cro2, cro1,cro3 );
    attValue(0) = MAL_S3_VECTOR_NORM ( cro3 );
    return attValue;
}




